"""The :class:`Pipeline` class is a composable a stream processor. It consists of a chain of
:class:`slurry.abc.Section`, which each handle a single stream processing operation.

The stream processing results are accessed by calling :meth:`Pipeline.tap` to create an
output channel. Each pipeline can have multiple open taps, each receiving a copy of the
output stream.

The pipeline can also be extended dynamically with new pipeline sections with
:meth:`Pipeline.extend`, adding additional processing.
"""
__version__ = '0.1.0'

from itertools import chain
import math
from typing import AsyncContextManager, Sequence

import trio
from async_generator import aclosing, asynccontextmanager

from .abc import Section
from .tap import Tap

class Pipeline:
    """The main Slurry ``Pipeline`` class.

    .. note::
        Do not instantiate a ``Pipeline`` class manually. Use :meth:`create`
        instead. It returns an async context manager which manages the pipeline lifetime.
    """
    def __init__(self, *sections: Sequence[Section],
                 main_nursery: trio.Nursery,
                 main_switch: trio.Event):
        self.sections = sections
        self._taps = set()
        self._main_nursery = main_nursery
        self._main_switch = main_switch

    @classmethod
    @asynccontextmanager
    async def create(cls, *sections: Sequence[Section]) -> AsyncContextManager["Pipeline"]:
        """Creates a new pipeline context and adds the given section sequence to it.

        :param sections: One or more pipeline sections.
            It is valid to supply an async iterable instead of a :class:`Section` as *first*
            section.
        :type sections: Sequence[slurry.abc.Section]
        """
        async with trio.open_nursery() as nursery:
            pipeline = cls(*sections, main_nursery=nursery, main_switch=trio.Event())
            nursery.start_soon(pipeline._pump) # pylint: disable=protected-access
            yield pipeline
            nursery.cancel_scope.cancel()

    async def _pump(self):
        """Runs the pipeline."""
        await self._main_switch.wait()

        if isinstance(self.sections[0], Section):
            first_input = None
            sections = self.sections
        else:
            first_input = self.sections[0]
            sections = self.sections[1:]

        channels = chain((first_input,), *(trio.open_memory_channel(0) for _ in sections))

        async with trio.open_nursery() as nursery:

            # Start pumps
            for section in sections:
                nursery.start_soon(section.pump, next(channels), next(channels))

            # Output to taps
            async with aclosing(next(channels)) as aiter:
                async for item in aiter:
                    for tap in self._taps:
                        nursery.start_soon(tap.send, item)

        # There is no more output to send. Close the taps.
        for tap in self._taps:
            await tap.send_channel.aclose()

    def tap(self, *,
            max_buffer_size: int = 0,
            timeout: float = math.inf,
            retrys: int = 0,
            start: bool = True) -> trio.MemoryReceiveChannel:
        # pylint: disable=line-too-long
        """Create a new output channel for this pipeline.

        Multiple channels can be opened and will receive a copy of the output data.

        .. note::
            The output is sent by reference, so if the output is a mutable type and
            a consumer changes it, other consumers will see the changed output.

        :param max_buffer_size: Although not recommended in general, it is possible to
            set a buffer on the output channel. (default ``0``) See
            `Buffering in channels <https://trio.readthedocs.io/en/stable/reference-core.html#buffering-in-channels>`_
            for further advice.
        :type max_buffer_size: int
        :param timeout: Timeout in seconds when attempting to send an item. (default ``math.inf``)
        :type timeout: float
        :param retrys: Number of times to retry sending, if the initial attempt fails.
            (default ``0``)
        :type retrys: int
        :param start: Start processesing when opening this tap. (default ``True``)
        :type start: bool

        :return: A trio ``MemoryReceiveChannel`` from which pipeline output can be pulled.
        """
        send_channel, receive_channel = trio.open_memory_channel(max_buffer_size)
        self._taps.add(Tap(send_channel, timeout, retrys))
        if start:
            self._main_switch.set()
        return receive_channel

    def extend(self, *sections: Sequence[Section], start: bool = False) -> "Pipeline":
        """Extend this pipeline into a new pipeline.

        :param sections: One or more pipeline sections.
        :type sections: Sequence[Section]
        :param start: Start processing when adding this extension. (default: ``False``)
        :type start: bool
        """
        pipeline = Pipeline(
            self.tap(start=start),
            sections,
            main_nursery=self._main_nursery,
            main_switch=self._main_switch
        )
        self._main_nursery.start_soon(pipeline._pump) # pylint: disable=protected-access
        return pipeline



""" Abstract Base Classes for building pipeline sections. """
from abc import ABC, abstractmethod
from typing import Any, AsyncIterable, Optional

import trio

class Section(ABC):
    """Each pipeline section takes inputs from an async iterable, processes it and sends it to an
    output.

    A Section must implement the ``pump`` abstract method, which will be scheduled to run as a task
    by the pipeline.
    """

    @abstractmethod
    async def pump(self, input: Optional[AsyncIterable[Any]], output: trio.MemorySendChannel):
        """The pump method must contain the logic that iterates the input, processes the indidual
        items, and feeds results to the output.

        :param input: The input data feed. Will be ``None`` for the first ``Section``, as the first
            ``Section`` is expected to supply it's own input.
        :type input: Optional[AsyncIterable[Any]]
        :param output: The output memory channel where results are sent.
        :type output: trio.MemorySendChannel
        """



"""An async streaming data processing framework."""
__version__ = '0.3.0'

from .pipeline import Pipeline

from .abc import Section
from .buffers import Window, Group, Delay, RateLimit
from .combiners import Chain, Merge, Zip, ZipLatest
from .filters import Skip, Filter, Changes
from .refiners import Map



"""Pipeline sections that produce data streams."""
from typing import Any

import trio
from async_generator import aclosing

from .abc import Section

class Repeat(Section):
    """Yields a single item repeatedly at regular intervals.

    If used as a middle section, the input can be used to set the value that is sent. When
    an input is received, it is sent immidiately, and the internal timer resets.

    :param interval: Delay between each transmission.
    :type interval: float
    :param default: Default item to send. If not supplied, will wait for an input.
    :type default: Any

    :raises RuntimeError: If used as a first section and no default is provided.
    """
    def __init__(self, interval: float, **kwargs):
        self.interval = interval
        self.kwargs = kwargs
        self.item = None
        self._item_set = trio.Event()

    async def pump(self, input, output):
        if input is None and 'default' not in self.kwargs:
            # pylint: disable=line-too-long
            raise RuntimeError('A default value must be provided, if Repeat is used as first section.')
        timer_cancel_scope = None

        async def set_item_task():
            async with aclosing(input) as aiter:
                async for item in aiter:
                    self.item = item
                    self._item_set.set()
                    if timer_cancel_scope is not None:
                        timer_cancel_scope.cancel()

        async with trio.open_nursery() as nursery:
            if input is not None:
                nursery.start_soon(set_item_task)
            if 'default' in self.kwargs:
                self.item = self.kwargs['default']
            else:
                await self._item_set.wait()
            while True:
                await output.send(self.item)
                with trio.CancelScope() as timer_cancel_scope:
                    await trio.sleep(self.interval)
